//	GamePlayState.cpp - The main gameplay state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "ScreamingSnakeballGlobalData.h"
#include "SnakeballSounder.h"
#include "drawing/Drawer.h"
#include "GamePlayState.h"
#include "Application.h"
#include "Random.h"
#include "Player.h"
#include "Ball.h"
#include "Wall.h"

#include <sstream>
#include <iomanip>

//------------------------------------------------------------------------------
GameStateCreator<GamePlayState> GamePlayState::creator;

//------------------------------------------------------------------------------
GamePlayState::GamePlayState(Application& app, SounderPtr& sound):
GameState(app, sound),
backgroundColour(0.125f),
backgroundTweener(backgroundColour, 0.125f, 0.2f),
borderColour(0.1875f),
borderTweener(borderColour, 0.1875f, 0.2f),
shake(0.0f),
shakeIndex(0.0f),
lineLeft(1.0f, 1.0f, 0.33f),
lineLeftTweener(lineLeft, ThreeFloats(0.33f, 0.33f, 1.0f), 0.005f),
lineRight(1.0f, 0.33f, 0.33f),
lineRightTweener(lineRight, ThreeFloats(0.33f, 1.0f, 0.33f), 0.005f),
leftScore(0),
rightScore(0),
lastTicks(0),
remainingTime(5.0f*60.0f),
goalPause(0),
redMeter(TwoFloats(62.0f, 8.0f), 1.0f),
blueMeter(TwoFloats(961.0f, 8.0f), -1.0f),
redPower(0.0f),
bluePower(0.0f),
lastWhisper(0),
sounder(dynamic_pointer_cast<SnakeballSounder>(sound))
{
	int i;
	const TwoFloats drawSize(getDrawingAreaSize());

	globalData = dynamic_pointer_cast<ScreamingSnakeballGlobalData>(app.getGlobalData());

	whispering = globalData->getWhispering();
	shouting = globalData->getShouting();
	remainingTime = globalData->getMatchLength() * 60.0f;

	sounder->setShouting(shouting);

	//Set up whispers.
	for(i=0;i<30;++i)
	{
		std::wstringstream tempstr;

		tempstr << L"Whisper" << std::setw(2) << std::setfill(L'0') << i;
		whispers.push_back(tempstr.str());
	}

	//Set up gradient lines.
	linePos[0].push_back(TwoFloats(64.0f+10.0f, 64.0f));
	linePos[0].push_back(TwoFloats(64.0f+10.0f, 64.0f));
	linePos[0].push_back(TwoFloats(64.0f+10.0f, 64.0f));
	linePos[0].push_back(TwoFloats(drawSize.x-(64.0f+12.0f), 64.0f));
	linePos[0].push_back(TwoFloats(drawSize.x-(64.0f+10.0f), 64.0f));
	linePos[0].push_back(TwoFloats(drawSize.x-(64.0f+10.0f), 64.0f));

	linePos[1].push_back(TwoFloats(64.0f+10.0f, drawSize.y-32.0f));
	linePos[1].push_back(TwoFloats(64.0f+10.0f, drawSize.y-32.0f));
	linePos[1].push_back(TwoFloats(64.0f+10.0f, drawSize.y-32.0f));
	linePos[1].push_back(TwoFloats(drawSize.x-(64.0f+12.0f), drawSize.y-32.0f));
	linePos[1].push_back(TwoFloats(drawSize.x-(64.0f+10.0f), drawSize.y-32.0f));
	linePos[1].push_back(TwoFloats(drawSize.x-(64.0f+10.0f), drawSize.y-32.0f));

	lineSize.push_back(4.0f);
	lineSize.push_back(4.0f);
	lineSize.push_back(4.0f);
	lineSize.push_back(4.0f);
	lineSize.push_back(4.0f);
	lineSize.push_back(4.0f);
	
	lineCol.push_back(lineLeft);
	lineCol.push_back(lineLeft);
	lineCol.push_back(lineLeft);
	lineCol.push_back(lineRight);
	lineCol.push_back(lineRight);
	lineCol.push_back(lineRight);

	lineAlpha.push_back(1.0f);
	lineAlpha.push_back(1.0f);
	lineAlpha.push_back(1.0f);
	lineAlpha.push_back(1.0f);
	lineAlpha.push_back(1.0f);
	lineAlpha.push_back(1.0f);

	//Setup the players.
	players.push_back(shared_ptr<Player>(new Player(TwoFloats(384.0f, 64.0f+((drawSize.y-96.0f)*0.5f)),
													1.5f * NIALL_PI,
													ThreeFloats(1.0f, 0.45f, 0.45f))));
	collidables.push_back(players.back());
	players.push_back(shared_ptr<Player>(new Player(TwoFloats(drawSize.x-384.0f, 64.0f+((drawSize.y-96.0f)*0.5f)),
													0.5f * NIALL_PI,
													ThreeFloats(0.53f, 0.45f, 1.0f))));
	collidables.push_back(players.back());

	//Add the ball.
	balls.push_back(shared_ptr<Ball>(new Ball(TwoFloats(drawSize.x*0.5f,
														64.0f+((drawSize.y-96.0f)*0.5f)))));
	collidables.push_back(balls.back());

	//Create the Collidable we use to represent the arena walls.
	arenaWall = shared_ptr<Wall>(new Wall(0.0f, L"Wall"));
	arenaWall->setRadius(18.0f);
	tailWall = shared_ptr<Wall>(new Wall(0.0f, L"TailWall"));
	tailWall->setRadius(20.0f);

	///	Setup the initial announcement.
	initialAnnouncement.push_back(L"3...");
	initialColours.push_back(ThreeFloats(1.0f, 0.45f, 0.45f));
	initialAnnouncement.push_back(L"2...");
	initialColours.push_back(ThreeFloats(1.0f, 0.45f, 0.8f));
	initialAnnouncement.push_back(L"1...");
	initialColours.push_back(ThreeFloats(0.53f, 0.45f, 1.0f));
	initialAnnouncement.push_back(L"Play ball!");
	initialColours.push_back(ThreeFloats(1.0f, 0.69f, 0.45f));
	announcer.trigger(L"Match starts in...", ThreeFloats(1.0f, 0.95f, 0.45f));
}

//------------------------------------------------------------------------------
GamePlayState::~GamePlayState()
{

}

//------------------------------------------------------------------------------
void GamePlayState::update()
{
	uint32_t ticks;
	unsigned int i;
	const TwoFloats drawSize(getDrawingAreaSize());
	const TwoFloats arenaTopLeft(64.0f, 64.0f);
	const TwoFloats arenaBottomRight(drawSize.x-64.0f, drawSize.y-32.0f);

	backgroundColLimit = 1;

	//--Update anything not related to the game mechanics-----------------------
	//Update any tweeners.
	if(!backgroundTweener.getFinished())
		backgroundTweener.tick();
	if(!borderTweener.getFinished())
		borderTweener.tick();
	if(!lineRightTweener.getFinished())
		lineRightTweener.tick();
	else
	{
		ThreeFloats tempCol;

		while((tempCol == 0.0f) || (tempCol == 1.0f) || (tempCol == lineRight))
		{
			tempCol.x = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.33f;
			tempCol.y = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.33f;
			tempCol.z = (Random::getGlobal().getRandomFloat() > 0.5f) ? 1.0f : 0.33f;
		}
		lineRightTweener.reset(tempCol);
	}
	if(!lineLeftTweener.getFinished())
		lineLeftTweener.tick();
	else
		lineLeftTweener.reset(lineRight);

	//Update screen shake.
	if(shake > 0.0f)
	{
		shake -= 0.05f;
		shakeIndex += 0.1f;
		if(shakeIndex > 1.0f)
			shakeIndex -= 1.0f;
	}

	if(!announcer.update())
	{
		if(initialAnnouncement.size() > 0)
		{
			announcer.trigger(initialAnnouncement[0], initialColours[0]);
			initialAnnouncement.erase(initialAnnouncement.begin());
			initialColours.erase(initialColours.begin());

			if(initialAnnouncement.size() == 0)
				sounder->triggerSound(L"BeepHi");
			else
				sounder->triggerSound(L"BeepLo");
		}
	}

	//If the game is finished, don't update anything.
	if((remainingTime <= 0.0f) || (initialAnnouncement.size() > 0))
		return;

	//--Update actual game stuff------------------------------------------------
	//Update remainingTime.
	if(!lastTicks)
		lastTicks = getApplication().getTimeSinceStart();
	else if(!goalPause)
	{
		ticks = getApplication().getTimeSinceStart();

		remainingTime -= (float)(ticks-lastTicks)/1000.0f;
		lastTicks = ticks;
		if(remainingTime < 0.0f)
		{
			remainingTime = 0.0f;
			if(leftScore > rightScore)
			{
				announcer.trigger(L"Red wins!", ThreeFloats(1.0f, 0.45f, 0.45f));
				sounder->triggerSound(L"PointScoredRed");
			}
			else if(rightScore > leftScore)
			{
				announcer.trigger(L"Blue wins!", ThreeFloats(0.53f, 0.45f, 1.0f));
				sounder->triggerSound(L"PointScoredBlue");
			}
			else
			{
				announcer.trigger(L"Humiliating draw!", ThreeFloats(1.0f, 0.69f, 0.45f));
				sounder->triggerSound(L"PointScoredRed");
				sounder->triggerSound(L"PointScoredBlue");
			}
		}
	}

	//Update the particles.
	particles.update();

	///Update the power levels.
	if(!whispering && !shouting)
	{
		float lastRed = redPower;
		float lastBlue = bluePower;

		redPower += (sounder->getRedInput() * 0.0033f);
		if(redPower > 1.0f)
			redPower = 1.0f;
		bluePower += (sounder->getBlueInput() * 0.0033f);
		if(bluePower > 1.0f)
			bluePower = 1.0f;

		if(redPower < 0.87f)
			redPower -= 0.0005f;
		else
			redPower -= 0.001f;
		if(redPower < 0.0f)
			redPower = 0.0f;
		if(bluePower < 0.87f)
			bluePower -= 0.0005f;
		else
			bluePower -= 0.001f;
		if(bluePower < 0.0f)
			bluePower = 0.0f;

		redMeter.setValue(redPower);
		blueMeter.setValue(bluePower);

		players[0]->setPowerLevel(redPower);
		players[1]->setPowerLevel(bluePower);

		players[0]->setTailFlash(bluePower >= 0.87f);
		players[1]->setTailFlash(redPower >= 0.87f);

		if((lastRed < 0.87f) && (redPower >= 0.87f))
			sounder->triggerSound(L"PowerUpRed");
		else if((lastRed >= 0.87f) && (redPower < 0.87f))
			sounder->triggerSound(L"PowerDownRed");
		if((lastBlue < 0.87f) && (bluePower >= 0.87f))
			sounder->triggerSound(L"PowerUpBlue");
		else if((lastBlue >= 0.87f) && (bluePower < 0.87f))
			sounder->triggerSound(L"PowerDownBlue");
	}

	if(shouting)
	{
		redMeter.setValue(sounder->getRedInput());

		players[0]->setTailFlash(sounder->getRedInput() >= 0.87f);
		players[1]->setTailFlash(sounder->getRedInput() >= 0.87f);
	}

	if(!goalPause)
	{
		//If we're playing Shouting Snakeball, update the ball's speed from the
		//red audio input.
		if(shouting)
		{
			for(i=0;i<balls.size();++i)
				balls[i]->updateSpeed(sounder->getRedInput());
		}

		//Update the players.
		for(i=0;i<players.size();++i)
			players[i]->update();
	}

	//Update the ball.
	for(i=0;i<balls.size();++i)
		balls[i]->update();

	//Do the collision detection.
	collisionDetection();

	//Make sure the ball stays within the arena.
	for(i=0;i<balls.size();++i)
		balls[i]->arenaWallCheck(drawSize);

	//Make sure the players stay within the arena.
	for(i=0;i<players.size();++i)
		players[i]->arenaWallCheck(drawSize);

	if(goalPause > 0)
	{
		--goalPause;
		if(!goalPause)
		{
			players[0]->reset(TwoFloats(384.0f, 64.0f+((drawSize.y-96.0f)*0.5f)),
							  1.5f * NIALL_PI);
			players[1]->reset(TwoFloats(drawSize.x-384.0f, 64.0f+((drawSize.y-96.0f)*0.5f)),
							  0.5f * NIALL_PI);
			balls[0]->reset(TwoFloats(drawSize.x*0.5f, 64.0f+((drawSize.y-96.0f)*0.5f)));

			lastTicks = getApplication().getTimeSinceStart();

			redPower = 0.0f;
			bluePower = 0.0f;
		}
	}
	else
	{
		//Check if one of the players has scored.
		for(i=0;i<balls.size();++i)
		{
			const TwoFloats& ballPos(balls[i]->getPosition());

			if(ballPos.x < arenaTopLeft.x)
			{
				++rightScore;
				rightScoreLayout.clear();
				goalPause = 120;
				balls[i]->setSlow();
				announcer.trigger(L"Blue point!", ThreeFloats(0.53f, 0.45f, 1.0f));
				sounder->triggerSound(L"PointScoredBlue");

				if(whispering || shouting)
				{
					players[0]->incTail(24);
					players[1]->incTail(-24);
				}
			}
			else if(ballPos.x > arenaBottomRight.x)
			{
				++leftScore;
				leftScoreLayout.clear();
				goalPause = 120;
				balls[i]->setSlow();
				announcer.trigger(L"Red point!", ThreeFloats(1.0f, 0.45f, 0.45f));
				sounder->triggerSound(L"PointScoredRed");

				if(whispering || shouting)
				{
					players[0]->incTail(-32);
					players[1]->incTail(32);
				}
			}
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::draw(DrawerPtr& d, float timestepInterp)
{
	unsigned int i;
	TwoFloats shakeVal;
	std::wstringstream tempstr;
	const TwoFloats& drawSize(d->getDrawingAreaSize());

	//--Arena background--------------------------------------------------------
	d->fillRect(0.0f, drawSize, backgroundColour);

	if(shake > 0.0f)
	{
		shakeVal.x = sinf(shakeIndex*2.0f*NIALL_PI) * 16.0f * shake * 0.25f * globalData->getScreenShake();
		d->translate(shakeVal);
		d->rotate(shakeVal.x * 0.001f, TwoFloats(640.0f, 360.0f));
	}

	//--Field markings.---------------------------------------------------------
	//Left goal line.
	d->drawTiledImage(L"VertLine",
					  TwoFloats(60.0f, 64.0f+(((drawSize.y-96.0f)*0.5f)-62.0f)),
					  TwoFloats(4.0f, 128.0f),
					  0.0f,
					  TwoFloats(1.0f, 128.0f/16.0f),
					  1.0f,
					  0.25f);
	//Right goal line.
	d->drawTiledImage(L"VertLine",
					  TwoFloats(drawSize.x-64.0f, 64.0f+(((drawSize.y-96.0f)*0.5f)-62.0f)),
					  TwoFloats(4.0f, 128.0f),
					  0.0f,
					  TwoFloats(1.0f, 128.0f/16.0f),
					  1.0f,
					  0.25f);
	//Centre line.
	d->drawTiledImage(L"VertLine",
					  TwoFloats((drawSize.x*0.5f)-2.0f, 64.0f),
					  TwoFloats(4.0f, drawSize.y-96.0f),
					  0.0f,
					  TwoFloats(1.0f, (drawSize.y-96.0f)/16.0f),
					  1.0f,
					  0.25f);

	//--Entities----------------------------------------------------------------

	//Draw the players' tails.
	for(i=0;i<players.size();++i)
		players[i]->drawTail(d, timestepInterp);
	//Draw the players.
	for(i=0;i<players.size();++i)
		players[i]->draw(d, timestepInterp);
	//Draw the balls.
	for(i=0;i<balls.size();++i)
		balls[i]->draw(d, timestepInterp, lineLeft, lineRight);

	particles.draw(d, timestepInterp);

	//--Border background-------------------------------------------------------
	//Top bar.
	d->fillRect(-64.0f, TwoFloats(drawSize.x+128.0f, 64.0f+64.0f), borderColour);
	//Bottom bar.
	d->fillRect(TwoFloats(-64.0f, drawSize.y-32.0f),
				TwoFloats(drawSize.x+128.0f, 32.0f+64.0f),
				borderColour);
	//Left bar top.
	d->fillRect(-64.0f,
				TwoFloats(64.0f+64.0f, 64.0f+((drawSize.y-96.0f)*0.5f)-64.0f+64.0f),
				borderColour);
	//Left bar bottom.
	d->fillRect(TwoFloats(-64.0f, 64.0f+((drawSize.y-96.0f)*0.5f)+64.0f),
				TwoFloats(64.0f+64.0f, (drawSize.y*0.5f)-64.0f+64.0f),
				borderColour);
	//Left bar middle.
	d->fillRect(TwoFloats(-64.0f, 64.0f+((drawSize.y-96.0f)*0.5f)-64.0f),
				TwoFloats(64.0f+22.0f, 128.0f),
				borderColour);
	//Right bar top.
	d->fillRect(TwoFloats(drawSize.x-64.0f, -64.0f),
				TwoFloats(64.0f+64.0f, 64.0f+((drawSize.y-96.0f)*0.5f)-64.0f+64.0f),
				borderColour);
	//Right bar bottom.
	d->fillRect(TwoFloats(drawSize.x-64.0f, 64.0f+((drawSize.y-96.0f)*0.5f)+64.0f),
				TwoFloats(64.0f+64.0f, (drawSize.y*0.5f)-64.0f+64.0f),
				borderColour);
	//Right bar middle.
	d->fillRect(TwoFloats(drawSize.x-22.0f, 64.0f+((drawSize.y-96.0f)*0.5f)-64.0f),
				TwoFloats(64.0f+22.0f, 128.0f),
				borderColour);

	//Left goal.
	d->drawImage(L"GoalGradient",
				 TwoFloats(22.0f, 64.0f+((drawSize.y-96.0f)*0.5f)-62.0f),
				 1.0f,
				 ThreeFloats(1.0f, 0.45f, 0.45f));
	//Right goal.
	d->drawImage(L"GoalGradient",
				 TwoFloats(drawSize.x-64.0f+40.0f, 64.0f+((drawSize.y-96.0f)*0.5f)-62.0f),
				 TwoFloats(-1.0f, 1.0f),
				 ThreeFloats(0.53f, 0.45f, 1.0f));

	//Border line.
	//Top.
	lineCol[0] = lineLeft;
	lineCol[1] = lineLeft;
	lineCol[2] = lineLeft;
	lineCol[3] = lineRight;
	lineCol[4] = lineRight;
	lineCol[5] = lineRight;
	d->drawConnectedLine(L"",
						 linePos[0],
						 lineSize,
						 lineCol,
						 lineAlpha);
	//Bottom.
	d->drawConnectedLine(L"",
						 linePos[1],
						 lineSize,
						 lineCol,
						 lineAlpha);
	//Left top.
	d->fillRect(TwoFloats(60.0f, 64.0f+10.0f),
				TwoFloats(4.0f, ((drawSize.y-96.0f)*0.5f)-64.0f-6.0f),
				lineLeft);
	//Left bottom.
	d->fillRect(TwoFloats(60.0f, 64.0f+((drawSize.y-96.0f)*0.5f)+64.0f-2.0f),
				TwoFloats(4.0f, ((drawSize.y-96.0f)*0.5f)-64.0f-6.0f),
				lineLeft);
	//Right top.
	d->fillRect(TwoFloats(drawSize.x-64.0f, 64.0f+10.0f),
				TwoFloats(4.0f, ((drawSize.y-96.0f)*0.5f)-64.0f-6.0f),
				lineRight);
	d->fillRect(TwoFloats(drawSize.x-64.0f, 64.0f+((drawSize.y-96.0f)*0.5f)+64.0f-2.0f),
				TwoFloats(4.0f, ((drawSize.y-96.0f)*0.5f)-64.0f-6.0f),
				lineRight);

	//Top-left Corner.
	d->drawImage(L"ArenaCornerBack",
				 TwoFloats(62.0f-4.0f, 64.0f-4.0f),
				 1.0f,
				 borderColour);
	//Top-right Corner.
	d->drawImage(L"ArenaCornerBack",
				 TwoFloats(drawSize.x-(62.0f-4.0f), 64.0f-4.0f),
				 TwoFloats(-1.0f, 1.0f),
				 borderColour);
	//Bottom-left Corner.
	d->drawImage(L"ArenaCornerBack",
				 TwoFloats(62.0f-4.0f, drawSize.y-(32.0f-4.0f)),
				 TwoFloats(1.0f, -1.0f),
				 borderColour);
	//Bottom-right Corner.
	d->drawImage(L"ArenaCornerBack",
				 TwoFloats(drawSize.x-(62.0f-4.0f), drawSize.y-(32.0f-4.0f)),
				 TwoFloats(-1.0f, -1.0f),
				 borderColour);

	//Top-left Corner.
	d->drawImage(L"ArenaCornerLine",
				 TwoFloats(62.0f-2.0f, 64.0f-2.0f),
				 1.0f,
				 lineLeft);
	//Top-right Corner.
	d->drawImage(L"ArenaCornerLine",
				 TwoFloats(drawSize.x-(62.0f-2.0f), 64.0f-2.0f),
				 TwoFloats(-1.0f, 1.0f),
				 lineRight);
	//Bottom-left Corner.
	d->drawImage(L"ArenaCornerLine",
				 TwoFloats(62.0f-2.0f, drawSize.y-(32.0f-2.0f)),
				 TwoFloats(1.0f, -1.0f),
				 lineLeft);
	//Bottom-right Corner.
	d->drawImage(L"ArenaCornerLine",
				 TwoFloats(drawSize.x-(62.0f-2.0f), drawSize.y-(32.0f-2.0f)),
				 TwoFloats(-1.0f, -1.0f),
				 lineRight);

	//Left goal line.
	d->drawImage(L"ArenaCornerBack",
				 TwoFloats(20.0f-2.0f, 64.0f+((drawSize.y-96.0f)*0.5f)-64.0f-2.0f),
				 1.0f,
				 borderColour);
	d->drawImage(L"ArenaCornerBack",
				 TwoFloats(20.0f-2.0f, 64.0f+((drawSize.y-96.0f)*0.5f)+64.0f+2.0f),
				 TwoFloats(1.0f, -1.0f),
				 borderColour);
	d->drawImage(L"GoalLine",
				 TwoFloats(20.0f, 64.0f+((drawSize.y-96.0f)*0.5f)-64.0f),
				 1.0f,
				 lineLeft);
	//Right goal line.
	d->drawImage(L"ArenaCornerBack",
				 TwoFloats(drawSize.x-20.0f+2.0f, 64.0f+((drawSize.y-96.0f)*0.5f)-64.0f-2.0f),
				 TwoFloats(-1.0f, 1.0f),
				 borderColour);
	d->drawImage(L"ArenaCornerBack",
				 TwoFloats(drawSize.x-20.0f+2.0f, 64.0f+((drawSize.y-96.0f)*0.5f)+64.0f+2.0f),
				 -1.0f,
				 borderColour);
	d->drawImage(L"GoalLine",
				 TwoFloats(drawSize.x-20.0f, 64.0f+((drawSize.y-96.0f)*0.5f)-64.0f),
				 TwoFloats(-1.0f, 1.0f),
				 lineRight);

	//--HUD---------------------------------------------------------------------
	//Left score.
	{
		tempstr.clear();
		tempstr.str(L"");
		tempstr << leftScore;
		d->drawText(tempstr.str(),
					L"ScoreFont",
					leftScoreLayout,
					TwoFloats(64.0f+4.0f, 168.0f+4.0f),
					(drawSize.x*0.5f)-(64.0f+4.0f)-16.0f,
					TextLayout::Right);
		d->drawText(tempstr.str(),
					L"ScoreFont",
					leftScoreLayout,
					TwoFloats(64.0f, 168.0f),
					(drawSize.x*0.5f)-64.0f-16.0f,
					TextLayout::Right,
					1.0f,
					ThreeFloats(1.0f, 0.45f, 0.45f));
	}
	//Right score.
	{
		tempstr.clear();
		tempstr.str(L"");
		tempstr << rightScore;
		d->drawText(tempstr.str(),
					L"ScoreFont",
					rightScoreLayout,
					TwoFloats((drawSize.x*0.5f)+16.0f+4.0f, 168.0f+4.0f),
					(drawSize.x*0.5f)-(64.0f+4.0f)-16.0f,
					TextLayout::Left);
		d->drawText(tempstr.str(),
					L"ScoreFont",
					rightScoreLayout,
					TwoFloats((drawSize.x*0.5f)+16.0f, 168.0f),
					(drawSize.x*0.5f)-64.0f-16.0f,
					TextLayout::Left,
					1.0f,
					ThreeFloats(0.53f, 0.45f, 1.0f));
	}

	//Remaining time.
	{
		tempstr.clear();
		tempstr.str(L"");
		tempstr << (int)(remainingTime/60.0f);
		tempstr << L":";
		tempstr << std::setw(2) << std::setfill(L'0') << ((int)remainingTime%60);
		d->drawText(tempstr.str(),
					L"TimeFont",
					TwoFloats(3.0f, 52.0f+3.0f),
					drawSize.x,
					TextLayout::Centre);
		d->drawText(tempstr.str(),
					L"TimeFont",
					TwoFloats(0.0f, 52.0f),
					drawSize.x,
					TextLayout::Centre,
					1.0f,
					(lineLeft + ((lineRight-lineLeft)*0.5f)));
	}

	//VU Meters.
	if(!whispering)
	{
		redMeter.draw(d);
		if(!shouting)
			blueMeter.draw(d);
	}

	//Announcer.
	announcer.draw(d);

	//--Reset screen shake------------------------------------------------------
	if(shake > 0.0f)
	{
		shakeVal.x = -shakeVal.x;
		d->translate(shakeVal);
		d->rotate(-shakeVal.x * 0.001f, TwoFloats(640.0f, 360.0f));
	}
}

//------------------------------------------------------------------------------
void GamePlayState::keyDown(const std::wstring& action,
							const std::wstring& keyPressed,
							int keyModifier)
{
	std::wstring tempstr;

	if(action == L"Escape")
		nextState(L"TitleScreen");
	else if(action == L"Restart")
		nextState(L"GamePlayState");
	else if(action != L"")
	{
		if(action[1] == L'1')
		{
			tempstr = action.substr(2);
			players[0]->keyDown(tempstr);
		}
		else if(action[1] == L'2')
		{
			tempstr = action.substr(2);
			players[1]->keyDown(tempstr);
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::keyUp(const std::wstring& action,
						  const std::wstring& keyPressed,
						  int keyModifier)
{
	std::wstring tempstr;

	if(action != L"")
	{
		if(action[1] == L'1')
		{
			tempstr = action.substr(2);
			players[0]->keyUp(tempstr);
		}
		else if(action[1] == L'2')
		{
			tempstr = action.substr(2);
			players[1]->keyUp(tempstr);
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::joyMove(const std::wstring& action,
							float value,
							int joystick,
							int axis)
{
	std::wstring tempstr;

	if(action != L"")
	{
		if(action[1] == L'1')
		{
			tempstr = action.substr(2);

			players[0]->move(tempstr, value);
		}
		else if(action[1] == L'2')
		{
			tempstr = action.substr(2);

			players[1]->move(tempstr, value);
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::triggerSFX(float level, float pan)
{
	int tempint;

	if(whispering || (shouting && (sounder->getRedInput() < 0.87f)))
	{
		tempint = lastWhisper;
		while(tempint == lastWhisper)
			tempint = Random::getGlobal().getRandomInt(whispers.size());
		sounder->triggerSound(whispers[tempint], level, pan);
		lastWhisper = tempint;
	}
	else
		sounder->triggerEnvelope(level, pan);
}

//------------------------------------------------------------------------------
void GamePlayState::collisionDetection()
{
	unsigned int i, j;

	for(i=0;i<collidables.size();++i)
	{
		shared_ptr<Collidable> colI(collidables[i].lock());

		if(colI)
		{
			const TwoFloats& iPos = colI->getPosition();
			const float iRad = colI->getRadius();

			arenaWallCheck(colI);
			if(whispering || (redPower < 0.87f))
				tailCheck(colI, players[1]);
			if(whispering || (bluePower < 0.87f))
				tailCheck(colI, players[0]);

			for(j=(i+1);j<collidables.size();++j)
			{
				shared_ptr<Collidable> colJ(collidables[j].lock());

				if(colJ)
				{
					const TwoFloats& jPos = colJ->getPosition();
					const float jRad = colJ->getRadius();

					if(colI->interestedInCollisionsWith(colJ->getName()))
					{
						if(iPos.getDistance(jPos) < (iRad+jRad))
						{
							if(!colJ->getIncorporeal())
								colI->collide(colJ, *this);

							if(!colI->getIncorporeal() &&
								(colJ->interestedInCollisionsWith(colI->getName())))
							{
								if(colJ->collide(colI, *this))
									triggerCollisionEffects(iPos + ((jPos-iPos)*0.5f));
							}
						}
					}
				}
			}
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::arenaWallCheck(shared_ptr<Collidable>& collidable)
{
	TwoFloats wallPos;
	const float wallRad = 18.0f;
	const float colRad = collidable->getRadius();
	const TwoFloats& colPos(collidable->getPosition());
	const TwoFloats arenaTopLeft(64.0f, 64.0f);
	const TwoFloats arenaBottomRight(getDrawingAreaSize().x-64.0f,
									 getDrawingAreaSize().y-32.0f);
	const float goalTop = 64.0f + ((getDrawingAreaSize().y-96.0f)*0.5f)-64.0f;
	const float goalBottom = goalTop + 128.0f;

	if((colPos.y > goalTop) && (colPos.y < goalBottom))
	{
		if((colPos.x-colRad) < (arenaTopLeft.x-42.0f))
		{
			wallPos.x = arenaTopLeft.x - 42.0f - wallRad;
			wallPos.y = colPos.y;
		}
		else if((colPos.x+colRad) > (arenaBottomRight.x+42.0f))
		{
			wallPos.x = arenaBottomRight.x + 42.0f + wallRad;
			wallPos.y = colPos.y;
		}
	}
	else
	{
		if((colPos.x-colRad) < arenaTopLeft.x)
		{
			wallPos.x = arenaTopLeft.x - wallRad;
			wallPos.y = colPos.y;
		}
		else if((colPos.x+colRad) > arenaBottomRight.x)
		{
			wallPos.x = arenaBottomRight.x + wallRad;
			wallPos.y = colPos.y;
		}
	}

	if((colPos.x < arenaTopLeft.x) || (colPos.x > arenaBottomRight.x))
	{
		if((colPos.y-colRad) < goalTop)
		{
			wallPos.y = goalTop - wallRad;
			if(wallPos.x == 0.0f)
				wallPos.x = colPos.x;
		}
		else if((colPos.y+colRad) > goalBottom)
		{
			wallPos.y = goalBottom + wallRad;
			if(wallPos.x == 0.0f)
				wallPos.x = colPos.x;
		}
	}
	else
	{
		if((colPos.y-colRad) < arenaTopLeft.y)
		{
			wallPos.y = arenaTopLeft.y - wallRad;
			if(wallPos.x == 0.0f)
				wallPos.x = colPos.x;
		}
		else if((colPos.y+colRad) > arenaBottomRight.y)
		{
			wallPos.y = arenaBottomRight.y + wallRad;
			if(wallPos.x == 0.0f)
				wallPos.x = colPos.x;
		}
	}

	if(wallPos != 0.0f)
	{
		float tempf = wallPos.getDistance(colPos);
		//Handle the corners.
		if(tempf > (wallRad+colRad))
			wallPos = colPos + ((wallPos-colPos).normal()) * (wallRad+colRad-5.0f);

		arenaWall->setPosition(wallPos);
		if(collidable->collide(arenaWall, *this))
			triggerCollisionEffects(collidable->getPosition());
	}
}

//------------------------------------------------------------------------------
void GamePlayState::tailCheck(shared_ptr<Collidable>& collidable,
							  const shared_ptr<Player>& player)
{
	unsigned int i;
	TwoFloats wallPos;
	const float wallRad = tailWall->getRadius();
	const float colRad(collidable->getRadius());
	const TwoFloats& colPos(collidable->getPosition());
	const std::vector<TwoFloats>& tail(player->getTail());
	shared_ptr<Player> p(dynamic_pointer_cast<Player>(collidable));
	std::vector<TwoFloats> intersectingSegments;

	if(p == player)
		return;

	for(i=8;i<tail.size();++i)
	{
		wallPos = tail[i-1] + ((tail[i]-tail[i-1])*0.5f);

		if(colPos.getDistance(wallPos) < (wallRad+colRad))
			intersectingSegments.push_back(wallPos);
	}

	if(intersectingSegments.size() > 0)
	{
		wallPos = 0.0f;
		for(i=0;i<intersectingSegments.size();++i)
			wallPos += intersectingSegments[i];
		wallPos /= (float)intersectingSegments.size();

		tailWall->setPosition(wallPos);
		collidable->collide(tailWall, *this);
	}
}

//------------------------------------------------------------------------------
void GamePlayState::triggerCollisionEffects(const TwoFloats& pos)
{
	int colourIndex;
	const float colInc = 1.0f/8.0f;
	const ThreeFloats lastBack(backgroundColour);
	const ThreeFloats lastBorder(borderColour);

	particles.trigger(pos);

	if(backgroundColLimit > 0)
	{
		colourIndex = Random::getGlobal().getRandomInt(3);
		switch(colourIndex)
		{
			case 0:
				backgroundColour.x += colInc;
				borderColour.y += colInc;
				borderColour.z += colInc;
				break;
			case 1:
				backgroundColour.y += colInc;
				borderColour.x += colInc;
				borderColour.z += colInc;
				break;
			case 2:
				backgroundColour.z += colInc;
				borderColour.x += colInc;
				borderColour.y += colInc;
				break;
			default:
				backgroundColour.x += colInc;
				borderColour.y += colInc;
				borderColour.z += colInc;
		}
		backgroundColour = lastBack + (globalData->getBackgroundFlash() * (backgroundColour-lastBack));
		borderColour = lastBorder + (globalData->getBorderFlash() * (borderColour-lastBorder));

		backgroundTweener.reset(ThreeFloats(0.125f));
		borderTweener.reset(ThreeFloats(0.1875f));

		--backgroundColLimit;
	}

	shake += 1.0f;
	if(shake > 2.0f)
		shake = 2.0f;
}
