//	OptionsScreen.h - The options screen state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef OPTIONSSCREEN_H_
#define OPTIONSSCREEN_H_

#include "drawing/TextLayout.h"
#include "OptionsSlider.h"
#include "ThreeFloats.h"
#include "GameState.h"
#include "VUMeter.h"
#include "Tweener.h"

class ScreamingSnakeballGlobalData;
class SnakeballSounder;

///	The options screen state.
class OptionsScreen : public GameState
{
  public:
	///	Constructor.
	OptionsScreen(Application& app, SounderPtr& sound);
	///	Destructor.
	~OptionsScreen();

	///	Called frequently to update the state.
	void update();
	///	Called to draw the state.
	void draw(DrawerPtr& d, float timestepInterp);

	///	Called when a key down event is received.
	void keyDown(const std::wstring& action,
				 const std::wstring& keyPressed,
				 int keyModifier);
	///	Called when a joystick axis changes.
	void joyMove(const std::wstring& action,
				 float value,
				 int joystick,
				 int axis);

	///	Returns the name of this state.
	const std::wstring getName() const {return L"OptionsScreen";};
	///	Used to construct instances of this state.
	static GameStateCreator<OptionsScreen> creator;
  private:
	///	Helper method. Moves/resizes any sliders that need moving.
	void updateSliderPositions();
	///	Helper method. Moves any text options which need moving.
	void updateTextPositions();
	///	Helper method. Updates the audio API.
	/*!
		\param dir True == right; false == left.
	 */
	void updateAudioAPI(bool dir);
	///	Helper method. Updates the red input text.
	void updateRedText();
	///	Helper method. Updates the blue input text.
	void updateBlueText();

	///	Struct representing a single entry in the menu.
	struct MenuEntry
	{
		///	The entry's position.
		TwoFloats position;
		///	The entry's colour.
		ThreeFloats colour;
		///	The entry's text.
		std::wstring text;
		///	The entry's font.
		std::wstring font;
		///	The entry's TextLayout.
		TextLayout layout;
	};

	///	The menu entries.
	/*!
		For simplicity (laziness) we store the positions for each entry for each
		of the four possible arrangements.
	 */
	MenuEntry entries[8][8];
	///	The four menu entry positions we use.
	TwoFloats entryPos[8];
	///	The position tweener for the first menu entry.
	STweener<TwoFloats> entryTween0;
	///	The position tweener for the second menu entry.
	STweener<TwoFloats> entryTween1;
	///	The position tweener for the third menu entry.
	STweener<TwoFloats> entryTween2;
	///	The position tweener for the fourth menu entry.
	STweener<TwoFloats> entryTween3;
	///	The position tweener for the fourth menu entry.
	STweener<TwoFloats> entryTween4;
	///	The position tweener for the fourth menu entry.
	STweener<TwoFloats> entryTween5;
	///	The position tweener for the fourth menu entry.
	STweener<TwoFloats> entryTween6;
	///	The position tweener for the fourth menu entry.
	STweener<TwoFloats> entryTween7;
	///	The currently-selected menu entry.
	int selectedEntry;
	///	Used by updateSliderPositions().
	int lastSelected;

	///	The screen shake slider.
	OptionsSlider screenShakeSlider;
	///	The background flash slider.
	OptionsSlider backgroundFlashSlider;
	///	The border flash slider.
	OptionsSlider borderFlashSlider;
	///	The match length slider.
	OptionsSlider matchLengthSlider;

	///	The audio API text.
	MenuEntry audioAPIText;
	///	The audio API text tweener.
	STweener<TwoFloats> audioAPITween;
	///	The red audio input text.
	MenuEntry redInputText;
	///	The red audio input text tweener.
	STweener<TwoFloats> redInputTween;
	///	The blue audio input text.
	MenuEntry blueInputText;
	///	The blue audio input text tweener.
	STweener<TwoFloats> blueInputTween;

	///	The red VU meter.
	VUMeter redMeter;
	///	The blue VU meter.
	VUMeter blueMeter;

	///	Used to handle menu navigation with a joystick.
	TwoFloats lastJoyVal;
	///	Used to handle menu navigation with a joystick.
	int joyRepeat;

	///	The current Audio API.
	std::wstring audioAPI;
	///	The current red audio input device.
	std::wstring redInput;
	///	The current blue audio input device.
	std::wstring blueInput;

	///	The index of the selected audio input device.
	int inputIndex;
	///	The red input channel.
	int redChannel;
	///	The blue input channel.
	int blueChannel;

	///	Our global data.
	shared_ptr<ScreamingSnakeballGlobalData> globalData;
	///	Our sounder.
	shared_ptr<SnakeballSounder> sounder;
};

#endif
